import typing as t

try:
    t.Literal
except AttributeError:
    from typing_extensions import Literal

    setattr(t, "Literal", Literal)

from pydantic import BaseModel, Json
from pydantic.generics import GenericModel


class Attributes(BaseModel):
    type: str
    url: str


class SalesforceRecord(BaseModel):
    attributes: Attributes


SalesforceRecordT = t.TypeVar("SalesforceRecordT", bound=SalesforceRecord)


class SalesforceRelationship(GenericModel, t.Generic[SalesforceRecordT]):
    totalSize: int
    done: bool
    records: t.List[SalesforceRecordT]


class GeneralEntitySchema(BaseModel):
    name: str
    value: t.Optional[str]


class Skill(BaseModel):
    name: str
    type: t.Optional[str]
    value: t.Optional[str]


class SalesforceExperience(SalesforceRecord):
    Certifications__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Company__c: t.Optional[str]
    Courses__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Date_Begin__c: t.Optional[str]
    Date_End__c: t.Optional[str]
    Description__c: t.Optional[str]
    Hash_Id__c: str
    Id__c: int
    Location_Fields__c: Json
    Location_Gmaps__c: t.Optional[str]
    Location_Lat__c: t.Optional[float]
    Location_Lng__c: t.Optional[float]
    Location_Text__c: str
    Profile__c: str
    Skills__c: t.Optional[Json[t.List[Skill]]]
    Tasks__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Title__c: t.Optional[str]


class SalesforceEducation(SalesforceRecord):
    Certifications__c: Json[t.List[GeneralEntitySchema]]
    Courses__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Date_Begin__c: t.Optional[str]
    Date_End__c: t.Optional[str]
    Description__c: t.Optional[str]
    Hash_Id__c: str
    Id__c: int
    Location_Fields__c: Json
    Location_Gmaps__c: t.Optional[str]
    Location_Lat__c: t.Optional[float]
    Location_Lng__c: t.Optional[float]
    Location_Text__c: str
    Profile__c: str
    School__c: t.Optional[str]
    Skills__c: t.Optional[Json[t.List[Skill]]]
    Tasks__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Title__c: t.Optional[str]


class SalesforceAttachment(SalesforceRecord):
    Alt__c: str
    Date_Edition__c: t.Optional[str]
    Extension__c: str
    File_Name__c: str
    File_Size__c: int
    Id__c: int
    Original_File_Name__c: str
    Profile__c: str
    Text__c: t.Optional[str]
    Timestamp__c: str
    Type__c: str
    URL__c: str


IntFlag = t.Literal[0, 1]
Gender = t.Literal["male", "M", "female", "F", "U", ""]
Seniority = t.Literal["senior", "junior", "dev"]
Status = t.Literal["inactive"]


class URLs(BaseModel):
    from_resume: t.Optional[t.List[str]]
    linkedin: t.Optional[str]
    twitter: t.Optional[str]
    facebook: t.Optional[str]
    github: t.Optional[str]


class SalesforceHrFlowProfile(SalesforceRecord):
    HrFlow_Profile_Experiences__r: t.Optional[
        SalesforceRelationship[SalesforceExperience]
    ]
    HrFlow_Profile_Educations__r: t.Optional[
        SalesforceRelationship[SalesforceEducation]
    ]
    HrFlow_Profile_Attachments__r: t.Optional[
        SalesforceRelationship[SalesforceAttachment]
    ]
    Archive__c: t.Optional[str]
    Certifications__c: Json[t.List[GeneralEntitySchema]]
    Courses__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Dataset_Id__c: int
    Date_Birth__c: t.Optional[str]
    Date_Edition__c: t.Optional[str]
    Date_Reception__c: t.Optional[str]
    Educations_Duration__c: t.Optional[float]
    Email__c: t.Optional[str]
    Experiences_Duration__c: t.Optional[float]
    First_Name__c: t.Optional[str]
    Gender__c: t.Optional[Gender]
    Hash_Id__c: str
    Id__c: int
    Interests__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Labels__c: t.Optional[Json]
    Languages__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    LastModifiedDate: str
    Last_Name__c: t.Optional[str]
    Location_Fields__c: Json
    Location_Gmaps__c: t.Optional[str]
    Location_Lat__c: t.Optional[float]
    Location_Lng__c: t.Optional[float]
    Location_Text__c: str
    Metadatas__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Name__c: t.Optional[str]
    Phone__c: t.Optional[str]
    Picture__c: t.Optional[str]
    Reference__c: t.Optional[str]
    Seniority__c: t.Optional[Seniority]
    Skills__c: t.Optional[Json[t.List[Skill]]]
    Summary__c: t.Optional[str]
    Tags__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Tasks__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Text_Language__c: t.Optional[str]
    Text__c: t.Optional[str]
    URLs__c: t.Optional[Json[URLs]]


class Section(BaseModel):
    name: str
    title: str
    description: str


class RangeFloat(BaseModel):
    name: str
    value_min: float
    value_max: float
    unit: str


class RangeDate(BaseModel):
    name: str
    value_min: str
    value_max: str


class SalesforceHrFlowJob(SalesforceRecord):
    Archive__c: t.Optional[str]
    Benefits__c: t.Optional[str]
    Board_Id__c: int
    Certifications__c: Json[t.List[GeneralEntitySchema]]
    Courses__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Culture__c: t.Optional[str]
    Date_Edition__c: t.Optional[str]
    Hash_Id__c: str
    Id__c: int
    Interviews__c: t.Optional[str]
    Languages__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    LastModifiedDate: str
    Location_Fields__c: Json
    Location_Gmaps__c: t.Optional[str]
    Location_Lat__c: t.Optional[float]
    Location_Lng__c: t.Optional[float]
    Location_Text__c: str
    Metadatas__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Name__c: str
    Picture__c: t.Optional[str]
    Ranges_Date__c: t.List[RangeDate]
    Ranges_Float__c: t.List[RangeFloat]
    Reference__c: t.Optional[str]
    Requirements__c: t.Optional[str]
    Responsibilities__c: t.Optional[str]
    Sections__c: t.List[Section]
    Skills__c: t.Optional[Json[t.List[Skill]]]
    Slug__c: str
    Status__c: IntFlag
    Summary__c: t.Optional[str]
    Tags__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    Tasks__c: t.Optional[Json[t.List[GeneralEntitySchema]]]
    URL__c: t.Optional[str]



import json
import mimetypes
import typing as t
from datetime import date
from io import BytesIO
from logging import LoggerAdapter
from zipfile import ZipFile

import requests
import typing_extensions as te
from pydantic import Field, PositiveInt

from hrflow_connectors.core import (
    DataType,
    FieldType,
    ParametersModel,
    ReadMode,
    Warehouse,
    WarehouseReadAction,
    WarehouseWriteAction,
)

GRANT_TYPE = "client_credentials"
TOKEN_SCOPE = "MatchingIndexation"
TOKEN_SCOPE_FULL_RIGHT = "Customer"
LIMIT = 100
TIMEOUT = 10
JOBS_DEFAULT_MAX_READ = 100


class ReadProfilesParameters(ParametersModel):
    client_id: str = Field(
        ...,
        description="Client ID used to access TalentSoft API",
        repr=False,
        field_type=FieldType.Auth,
    )
    client_secret: str = Field(
        ...,
        description="Client Secret used to access TalentSoft API",
        repr=False,
        field_type=FieldType.Auth,
    )
    client_url: str = Field(
        ...,
        description="URL of TalentSoft client integration",
        field_type=FieldType.Other,
    )
    filter: t.Optional[str] = Field(
        None,
        description=(
            "Filter to apply when reading profiles. See documentation at"
            " https://developers.cegid.com/api-details#api=cegid-talentsoft-recruiting-"
            "matchingindexation&operation=api-exports-v1-candidates-get ."
            " Examples : By id Single Item 'id::_TS-00001'; By id Multiple"
            " Items 'id::_TS-00001,_TS-00002'; By email"
            " 'email::john.doe@company.corp'; By updateDate updated before the 10th"
            " of June 2019 'updateDate:lt:2019-06-10'; By chronoNumber greater than"
            " 108921  'chronoNumber:gt:108921'"
        ),
        field_type=FieldType.QueryParam,
    )
    fileId: t.Optional[str] = Field(
        description=(
            "If provided only the attachment matching with fileId is left in"
            " 'attachments'. If not found all attachments are left."
        ),
        field_type=FieldType.QueryParam,
    )
    only_resume: bool = Field(
        False,
        description="If enabled only resume attachments are returned",
        field_type=FieldType.QueryParam,
    )


class ReadJobsParameters(ParametersModel):
    client_id: str = Field(
        ...,
        description="Client ID used to access TalentSoft API",
        repr=False,
        field_type=FieldType.Auth,
    )
    client_secret: str = Field(
        ...,
        description="Client Secret used to access TalentSoft API",
        repr=False,
        field_type=FieldType.Auth,
    )
    client_url: str = Field(
        ...,
        description="URL of TalentSoft client integration",
        field_type=FieldType.Other,
    )
    q: t.Optional[str] = Field(
        None,
        description="Query search to get vacancies",
        field_type=FieldType.QueryParam,
    )
    filter: t.Optional[str] = Field(
        None,
        description=(
            "Filter to apply when reading vacancies. See documentation at"
            " https://developers.cegid.com/api-details#api=cegid-talentsoft-recruiting"
            "-matchingindexation&operation=api-exports-v1-vacancies-get"
            " . . You can filter by **chronoNumber**, **updateDate**, **reference**"
            " **vacancyStatus**, **clientVacancyStatus**, **publicationMedia** "
            " **publishedOnTheMedia**. Examples : By reference Single Item"
            " 'reference::2019-01'; By reference Multiple Items"
            " 'reference::2019-01,2019-02,2019-03';  By updateDate updated before the"
            " 10th of June 2019 'updateDate:lt:2019-06-10'; By chronoNumber greater"
            " than 108921  'chronoNumber:gt:108921' . "
        ),
        field_type=FieldType.QueryParam,
    )
    max_read: t.Optional[PositiveInt] = Field(
        JOBS_DEFAULT_MAX_READ,
        description=(
            "The maximum number of jobs to pull during the execution. Proper tuning of"
            " this parameter should allow to control the execution time and avoid"
            " overtimes"
        ),
        field_type=FieldType.Other,
    )


class WriteProfileParameters(ParametersModel):
    client_id: str = Field(
        ...,
        description="client id used to access TalentSoft front office API",
        repr=False,
        field_type=FieldType.Auth,
    )
    client_secret: str = Field(
        ...,
        description="client secret used to access TalentSoft front office API",
        repr=False,
        field_type=FieldType.Auth,
    )
    client_url: str = Field(
        ...,
        description="url used to access TalentSoft front office API",
        repr=False,
        field_type=FieldType.Auth,
    )
    job_reference: str = Field(
        None,
        description="reference of the job offer to which the candidate is applying",
        repr=False,
        field_type=FieldType.Auth,
    )


def decode_unicode(input_str: str) -> str:
    try:
        return bytes(input_str, "utf-8").decode("unicode_escape")
    except UnicodeDecodeError:
        return input_str


def decode_json(
    obj: t.Union[str, list, dict, t.Any]
) -> t.Union[str, list, dict, t.Any]:
    if isinstance(obj, str):
        return decode_unicode(obj)
    elif isinstance(obj, list):
        return [decode_json(item) for item in obj]
    elif isinstance(obj, dict):
        return {key: decode_json(value) for key, value in obj.items()}
    else:
        return obj


def get_mime_type(filename: t.Optional[str]) -> str:
    if filename is None:
        return "application/octet-stream"
    mime_type, encoding = mimetypes.guess_type(filename)
    return mime_type or "application/octet-stream"


def get_cv_content(attachment: dict) -> t.Optional[bytes]:
    response = requests.get(attachment["public_url"])
    if response.status_code == 200:
        return response.content
    raise Exception(response.text)


def get_talentsoft_auth_token(
    client_url: str,
    client_id: str,
    client_secret: str,
    scope: str = TOKEN_SCOPE,
    front_or_back: te.Literal["back", "front"] = "back",
) -> str:
    if front_or_back == "front":
        data = dict(
            grant_type=GRANT_TYPE,
            client_id=client_id,
            client_secret=client_secret,
        )
    else:
        data = dict(
            grant_type=GRANT_TYPE,
            client_id=client_id,
            client_secret=client_secret,
            scope=scope,
        )
    response = requests.post(
        "{}/api/token".format(client_url),
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
        },
        timeout=TIMEOUT,
        data=data,
    )
    if not response.ok:
        raise Exception(
            "Failed to get authentication token with error={}".format(response.text)
        )
    try:
        return response.json()["access_token"]
    except (KeyError, requests.exceptions.JSONDecodeError) as e:
        raise Exception(
            "Failed to get token from response with error={}".format(repr(e))
        )


def post_applicant_front(client_url, token, applicant, files, job_reference=None):
    headers = {
        "Authorization": "Bearer " + token,
    }
    if job_reference:
        headers.update(
            {
                "Accept-Language": "fr-FR",
                "jobAdReference": job_reference,
            }
        )
    response = requests.post(
        "{}/api/v2/applicants/applicationswithoutaccount".format(client_url),
        headers=headers,
        data=applicant,
        files=files,
    )
    if response.status_code == 201:
        return response.json()
    raise Exception(response.text)


def read_jobs(
    adapter: LoggerAdapter,
    parameters: ReadProfilesParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    if parameters.filter and read_mode is ReadMode.incremental:
        raise Exception("filter cannot be set when read_mode is incremental")

    adapter.info("Requesting Authentication Token from TS")
    token = get_talentsoft_auth_token(
        client_url=parameters.client_url,
        client_id=parameters.client_id,
        client_secret=parameters.client_secret,
    )
    adapter.info("Authentication with TS API Endpoint finished")

    limit = min(LIMIT, parameters.max_read)
    params = dict(offset=0, limit=limit)
    if read_mode is ReadMode.incremental:
        if read_from is None:
            read_from = date.today().isoformat()
        params["filter"] = "updateDate:gt:{}".format(read_from)
    elif parameters.filter:
        params["filter"] = parameters.filter

    if parameters.q:
        params["q"] = parameters.q

    item_counter = 0
    while True:
        response = requests.get(
            "{}/api/exports/v1/vacancies".format(parameters.client_url),
            params=params,
            timeout=TIMEOUT,
            headers={
                "Authorization": "bearer {}".format(token),
            },
        )
        if not response.ok:
            raise Exception(
                "Failed to fetch jobs with params={} from TalentSoft with"
                " error={}".format(params, response.text)
            )
        if response.headers.get("Content-Length") == 0 or not response.content:
            if params["offset"] == 0:
                adapter.info(
                    "No jobs found with params={} text={} headers={}".format(
                        params, response.text, response.headers
                    )
                )
            return

        data = response.content
        while data:
            read_up_to = int.from_bytes(data[0:4], byteorder="little")
            zip_data = ZipFile(BytesIO(data[4 : 4 + read_up_to]))
            job = json.loads(zip_data.read("offerdetail").decode())["offerDetail"]
            yield job
            item_counter += 1
            if item_counter >= parameters.max_read:
                return
            data = data[4 + read_up_to :]
        params["offset"] += limit


def read_profiles(
    adapter: LoggerAdapter,
    parameters: ReadProfilesParameters,
    read_mode: t.Optional[ReadMode] = None,
    read_from: t.Optional[str] = None,
) -> t.Iterable[t.Dict]:
    token = get_talentsoft_auth_token(
        client_url=parameters.client_url,
        client_id=parameters.client_id,
        client_secret=parameters.client_secret,
    )
    params = dict(offset=0, limit=LIMIT)
    if parameters.filter:
        params["filter"] = parameters.filter
    if parameters.only_resume:
        params["attachmenttype"] = "resume"

    while True:
        response = requests.get(
            "{}/api/exports/v1/candidates".format(parameters.client_url),
            params=params,
            timeout=TIMEOUT,
            headers={
                "Authorization": "bearer {}".format(token),
            },
        )
        if not response.ok:
            raise Exception(
                "Failed to fetch candidates with params={} from TalentSoft with"
                " error={}".format(params, response.text)
            )
        if response.headers.get("Content-Length") == 0 or not response.content:
            if params["offset"] == 0:
                adapter.info(
                    "No profiles found with params={} text={} headers={}".format(
                        params, response.text, response.headers
                    )
                )
            return

        data = response.content
        while data:
            read_up_to = int.from_bytes(data[0:4], byteorder="little")
            zip_data = ZipFile(BytesIO(data[4 : 4 + read_up_to]))
            candidate = json.loads(zip_data.read("applicantdetail").decode())
            for attachment in candidate["attachments"]:
                attachment["raw"] = zip_data.read(attachment["id"])
            if parameters.fileId is not None:
                found = next(
                    (
                        attachment
                        for attachment in candidate["attachments"]
                        if attachment["id"] == parameters.fileId
                    ),
                    None,
                )
                if found:
                    candidate["attachments"] = [found]
            yield candidate
            data = data[4 + read_up_to :]
        params["offset"] += LIMIT


def write_profiles(
    adapter: LoggerAdapter,
    parameters: WriteProfileParameters,
    profiles: t.Iterable[t.Dict],
) -> t.List[t.Dict]:
    failed_profiles = []
    adapter.info("Requesting Authentication Token from TS")
    token = get_talentsoft_auth_token(
        client_url=parameters.client_url,
        client_id=parameters.client_id,
        client_secret=parameters.client_secret,
        front_or_back="front",
    )
    adapter.info("Authentication with TS API Endpoint finished")
    for profile in profiles:
        attachment = profile.pop("attachment", None)
        if not attachment:
            adapter.error("No attachment found for profile={}".format(profile))
            failed_profiles.append(profile)
            continue
        try:
            cv_content = get_cv_content(attachment)
        except Exception as e:
            adapter.error("Failed to get cv content with response={}".format(e))
            failed_profiles.append(profile)
            continue
        filename = attachment["original_file_name"]
        mime_type = get_mime_type(filename)
        files = [
            (
                "cv_file_id",
                (filename, cv_content, mime_type),
            )
        ]
        if parameters.job_reference:
            profile["application"]["offerReference"] = parameters.job_reference
        profile = decode_json(profile)
        profile_ts = dict(applicantApplication=json.dumps(profile))
        try:
            post_applicant_front(
                parameters.client_url,
                token,
                profile_ts,
                files,
                parameters.job_reference,
            )
        except Exception as e:
            adapter.error("Failed to write profile with response={}".format(e))
            failed_profiles.append(profile)

    return failed_profiles


TalentSoftProfilesWarehouse = Warehouse(
    name="TalentSoft Profiles",
    data_type=DataType.profile,
    read=WarehouseReadAction(
        parameters=ReadProfilesParameters,
        function=read_profiles,
    ),
    write=WarehouseWriteAction(
        parameters=WriteProfileParameters,
        function=write_profiles,
    ),
)


TalentSoftJobsWarehouse = Warehouse(
    name="TalentSoft Jobs",
    data_type=DataType.job,
    read=WarehouseReadAction(
        parameters=ReadJobsParameters,
        function=read_jobs,
        supports_incremental=True,
        item_to_read_from=lambda job: job["modificationDate"],
    ),
)



CIVILITY = [
    {
        "code": 3445,
        "clientCode": "_TS_CO_Civility_Dr",
        "label": "Dr.",
        "type": "civility",
        "parentType": "",
    },
    {
        "code": 1790,
        "clientCode": "_TS_CO_Civility_Miss",
        "label": "Miss",
        "type": "civility",
        "parentType": "",
    },
    {
        "code": 280,
        "clientCode": "_TS_CO_Civility_Mr",
        "label": "Mr.",
        "type": "civility",
        "parentType": "",
    },
    {
        "code": 1789,
        "clientCode": "_TS_CO_Civility_Mrs ",
        "label": "Mrs",
        "type": "civility",
        "parentType": "",
    },
    {
        "code": 281,
        "clientCode": "_TS_CO_Civility_Mme",
        "label": "Ms.",
        "type": "civility",
        "parentType": "",
    },
    {
        "code": 3467,
        "clientCode": "_TS_CO_Civility_Other",
        "label": "Other",
        "type": "civility",
        "parentType": "",
    },
]

EDUCATIONS_REFERENTIEL = {
    "Aucun diplôme": {
        "code": 261,
        "clientCode": "_TS_etude_min_Aucun_diplome",
        "label": "No Degree",
        "type": "educationLevel",
        "parentType": "",
    },
    "BAC": {
        "code": 262,
        "clientCode": "_TS_etude_min_BAC",
        "label": "Bac",
        "type": "educationLevel",
        "parentType": "",
    },
    "BAC+2": {
        "code": 2859,
        "clientCode": "_TS_etude_min_BAC2",
        "label": "A Levels, Diploma",
        "type": "educationLevel",
        "parentType": "",
    },
    "BAC+3": {
        "code": 4028,
        "clientCode": "_TS_etude_min_Licence",
        "label": "Licence",
        "type": "educationLevel",
        "parentType": "",
    },
    "BAC+4": {
        "code": 2860,
        "clientCode": "_TS_etude_min_BAC4",
        "label": "Bachelor's Degree",
        "type": "educationLevel",
        "parentType": "",
    },
    "BAC+5": {
        "code": 266,
        "clientCode": "_TS_etude_min_BAC5",
        "label": "Master Degree",
        "type": "educationLevel",
        "parentType": "",
    },
    "CAP, BEP": {
        "code": 2858,
        "clientCode": "_TS_etude_min_CAP_BEP",
        "label": "GCSEs",
        "type": "educationLevel",
        "parentType": "",
    },
    "DOCTORAT": {
        "code": 2861,
        "clientCode": "_TS_etude_min_DOCTORAT",
        "label": "PhD, Doctorate",
        "type": "educationLevel",
        "parentType": "",
    },
    "Mastère": {
        "code": 4030,
        "clientCode": "_TS_etude_min_Mastere",
        "label": "Mastère",
        "type": "educationLevel",
        "parentType": "",
    },
}
EXPERIENCES_REFERENTIEL = {
    "Etudiant-e": "",
    "Débutant-e/première expérience": {
        "code": 274,
        "clientCode": "_TS_niveau_exp_premiere_exp",
        "label": "First experience",
        "type": "experienceLevel",
        "parentType": "",
    },
    "Supérieure à 3 ans": {
        "code": 1785,
        "clientCode": "_TS_niveau_exp_superieur_3ans ",
        "label": "More than 3 years",
        "type": "experienceLevel",
        "parentType": "",
    },
    "Supérieure à 5 ans": {
        "code": 1783,
        "clientCode": "_TS_niveau_exp_superieur_5ans",
        "label": "More than 5 years",
        "type": "experienceLevel",
        "parentType": "",
    },
    "Supérieure à 8 ans": {
        "code": 275,
        "clientCode": "_TS_niveau_exp_superieur_8ans",
        "label": "More than 8 years",
        "type": "experienceLevel",
        "parentType": "",
    },
}

CONTRACT_TYPE_REFERENTIAL = {
    "Alternance": {
        "code": 1047,
        "clientCode": "_TS_CO_Contract_Alternance",
        "label": "Apprenticeship",
        "type": "contractType",
        "parentType": "",
    },
    "CDD": {
        "code": 1045,
        "clientCode": "_TS_CO_Contract_CDD",
        "label": "Fixed Term contract",
        "type": "contractType",
        "parentType": "",
    },
    "CDI": {
        "code": 1044,
        "clientCode": "_TS_CO_Contract_CDI",
        "label": "Permanent",
        "type": "contractType",
        "parentType": "",
    },
    "CIFRE": {
        "code": 1049,
        "clientCode": "_TS_CO_Contract_CIFRE",
        "label": "CIFRE",
        "type": "contractType",
        "parentType": "",
    },
    "Stage": {
        "code": 1046,
        "clientCode": "_TS_CO_Contract_Stage",
        "label": "Internship / Student",
        "type": "contractType",
        "parentType": "",
    },
    "VIE": {
        "code": 1048,
        "clientCode": "_TS_CO_Contract_VIE",
        "label": "VIE",
        "type": "contractType",
        "parentType": "",
    },
}



import typing as t

from pydantic import BaseModel


class PersonalInformation(BaseModel):
    civility: t.Optional[str]
    middleName: t.Optional[str]
    title: t.Optional[str]
    address: t.Optional[str]
    city: t.Optional[str]
    postalCode: t.Optional[str]
    birthDate: t.Optional[str]
    country: t.Optional[str]
    skypeAccount: t.Optional[str]
    receiveSMS: t.Optional[str]
    phoneNumber2: t.Optional[str]
    professionalEmail: t.Optional[str]
    sex: t.Optional[str]
    nationalities: t.List[str]
    frenchDisabledWorkerStatus: t.Optional[str]
    frenchPriorityNeighbourhood: t.Optional[str]
    firstName: str
    lastName: str
    email: str
    phoneNumber: str


class Diplomas(BaseModel):
    educationLevel: str
    diplomaCode: t.Optional[str]
    specialisation: t.Optional[str]
    yearObtained: str
    college: str
    collegeCity: t.Optional[str]


class Language(BaseModel):
    language: str
    languageLevel: str


class Education(BaseModel):
    diplomas: t.List[Diplomas]
    studiedLanguages: t.List[Language]


class Experience(BaseModel):
    experienceLevel: t.Optional[str]
    Profile: t.Optional[str]
    contract: t.Optional[str]
    company: str
    function: str
    length: t.Optional[str]


class Experiences(BaseModel):
    experienceLevel: t.Optional[str]
    experienceList: t.List[Experience]


class Mobility(BaseModel):
    geographicalAreas: t.List[str]
    countries: t.List[str]
    regions: t.List[str]
    departments: t.List[str]


class Availability(BaseModel):
    acceptsExtra: t.Optional[str]
    values: t.List[str]


class FurtherInformation(BaseModel):
    skills: t.List[str]


class eEOInformation(BaseModel):
    doesNotComplete: bool
    sex: t.Optional[str]
    race: t.Optional[str]
    ethnicity: t.Optional[str]
    veteranStatus: t.Optional[str]
    incapacityStatus: t.Optional[str]


class jobPreferences(BaseModel):
    primaryProfile: t.Optional[str]
    contract: t.Optional[str]
    contractDuration: t.Optional[str]
    salaryPretensions: t.Optional[str]
    dateOfAvailability: t.Optional[str]
    mobility: Mobility
    noticeDuration: t.Optional[str]
    mobilityDelay: t.Optional[str]
    trainingDateStart: t.Optional[str]
    trainingDateEnd: t.Optional[str]
    jobTime: t.Optional[str]
    secondaryProfiles: t.List[str]
    availability: Availability


class standardItem(BaseModel):
    code: int
    clientCode: str
    label: str
    active: bool
    parentCode: t.Union[None, int]
    type: str
    parentType: str
    hasChildren: bool


class fileItem(BaseModel):
    guid: str
    name: str
    description: str
    fileType: standardItem


class Application(BaseModel):
    id: int
    type: str
    offerReference: str
    offerTitle: str
    isOfferPublished: bool
    organisation: standardItem
    origin: standardItem
    motivation: t.Optional[str]
    referralCode: t.Optional[str]
    files: t.List[fileItem]
    applicationAnswers: t.List[dict]
    date: str
    status: standardItem
    personalDataConsentReceived: t.Optional[str]
    retentionDelay: t.Optional[str]
    frenchDisabledWorkerStatus: t.Optional[str]


class UploadedFile(BaseModel):
    description: str
    fileTypeId: str
    key: str


class Applicant(BaseModel):
    personalInformation: PersonalInformation
    jobPreferences: jobPreferences
    educations: Education
    experiences: Experiences
    consents: t.List[dict]
    furtherInformation: FurtherInformation
    eEOInformation: eEOInformation
    customFields: str


class TalentsoftApplicantSchema(BaseModel):
    applicant: Applicant
    application: Application
    uploadedFiles: t.List[UploadedFile]


class UpdateSpecialEmploymentRegulationsInFrance(BaseModel):
    disabledWorkerStatus: bool
    priorityNeighbourhood: bool


class UpdatePersonalInformation(BaseModel):
    birthDate: str
    nationalities: t.List[str]
    address: str
    postalCode: str
    city: str
    residentCountryId: str
    specialEmploymentRegulationsInFrance: t.Optional[
        UpdateSpecialEmploymentRegulationsInFrance
    ]


class UpdateJobPreferences(BaseModel):
    primaryProfileId: str
    dateOfAvailability: str
    salaryExpectations: str


class UpdateLanguage(BaseModel):
    languageId: str
    languageLevelId: str


class UpdateEducation(BaseModel):
    diplomaId: str
    educationLevelId: str


class UpdateExperience(BaseModel):
    experienceLevelId: str
    profileId: str
    company: str
    function: str
    contractTypeId: str


class UpdateAttachment(BaseModel):
    description: str
    key: str
    fileType: str


class CandidateUpdated(BaseModel):
    employeeNumber: t.Optional[str]
    lastName: str
    firstName: str
    middleName: t.Optional[str]
    email: str
    phoneNumber: t.Optional[str]
    civilityId: t.Optional[str]
    personalInformation: t.Optional[UpdatePersonalInformation]
    jobPreferences: t.Optional[UpdateJobPreferences]
    languages: t.Optional[t.List[UpdateLanguage]]
    educations: t.Optional[t.List[UpdateEducation]]
    experiences: t.Optional[t.List[UpdateExperience]]
    attachments: t.Optional[t.List[UpdateAttachment]]



import html
import re
import typing as t
from logging import LoggerAdapter

from hrflow import Hrflow
from pydantic import Field

from hrflow_connectors.connectors.hrflow.schemas import HrFlowJob
from hrflow_connectors.core import (
    DataType,
    FieldType,
    ParametersModel,
    Warehouse,
    WarehouseWriteAction,
)

LIST_JOBS_LIMIT = 30

LABEL_TO_JOB_FIELD = dict(
    course="courses",
    task="tasks",
    certification="certifications",
    language="languages",
)
SKILL_LABEL_TO_TYPE = dict(Skill=None, skill_hard="hard", skill_soft="soft")


class JobParsingException(Exception):
    def __init__(self, *args, client_response: t.Dict):
        self.client_response = client_response


class WriteJobParameters(ParametersModel):
    api_secret: str = Field(
        ...,
        description="X-API-KEY used to access HrFlow.ai API",
        repr=False,
        field_type=FieldType.Auth,
    )
    api_user: str = Field(
        ...,
        description="X-USER-EMAIL used to access HrFlow.ai API",
        field_type=FieldType.Auth,
    )
    board_key: str = Field(
        ..., description="HrFlow.ai board key", field_type=FieldType.QueryParam
    )
    sync: bool = Field(
        True,
        description="When enabled only pushed jobs will remain in the board",
        field_type=FieldType.Other,
    )
    update_content: bool = Field(
        False,
        description="When enabled jobs already present in the board are updated",
        field_type=FieldType.Other,
    )
    enrich_with_parsing: bool = Field(
        False,
        description="When enabled jobs are enriched with HrFlow.ai parsing",
        field_type=FieldType.Other,
    )


def remove_html_tags(text: str) -> str:
    return re.sub("<[^<]+?>", "", text)


def enrich_job_with_parsing(hrflow_client: Hrflow, job: t.Dict) -> None:
    concatenate = []
    summary = job.get("summary")
    if summary:
        concatenate.append(summary)

    for section in job.get("sections") or []:
        description = section.get("description")
        if description:
            concatenate.append(description)

    concatenated = "\n".join(concatenate)
    cleaned = html.unescape(remove_html_tags(concatenated)).strip()
    if cleaned == "":
        return

    response = hrflow_client.document.parsing.post(text=cleaned)
    if response["code"] >= 400:
        raise JobParsingException("Failed to parse job", client_response=response)

    entities, parsed_text = response["data"]["entities"], response["data"]["text"]
    for field in ["skills", "languages", "certifications", "courses", "tasks"]:
        if job.get(field) is None:
            job[field] = []

    for entitiy in entities:
        label = entitiy["label"]
        entity_text = parsed_text[entitiy["start"] : entitiy["end"]]

        if label in LABEL_TO_JOB_FIELD:
            job_field = LABEL_TO_JOB_FIELD[label]
            if next(
                (
                    element
                    for element in job[job_field]
                    if element["name"] == entity_text
                ),
                False,
            ):
                continue
            job[job_field].append(dict(name=entity_text, value=None))
        elif label in SKILL_LABEL_TO_TYPE:
            skill_type = SKILL_LABEL_TO_TYPE[label]
            if next(
                (
                    skill
                    for skill in job["skills"]
                    if skill["name"] == entity_text and skill["type"] == skill_type
                ),
                False,
            ):
                continue
            job["skills"].append(dict(name=entity_text, type=skill_type, value=None))

    return


def write(
    adapter: LoggerAdapter, parameters: WriteJobParameters, jobs: t.Iterable[t.Dict]
) -> t.List[t.Dict]:
    failed_jobs = []
    hrflow_client = Hrflow(
        api_secret=parameters.api_secret, api_user=parameters.api_user
    )
    if parameters.sync is True:
        jobs = list(jobs)
        references_to_push = set(
            [job.get("reference") for job in jobs if job.get("reference")]
        )

        references_in_board = set()
        page = 1
        while True:
            response = hrflow_client.job.storing.list(
                board_keys=[parameters.board_key], limit=LIST_JOBS_LIMIT, page=page
            )
            if response["code"] >= 400:
                adapter.error(
                    "Failed to list jobs in board board_key={} "
                    "limit={} page={} response={}".format(
                        parameters.board_key, LIST_JOBS_LIMIT, page, response
                    )
                )
                raise Exception("Failed to list jobs in board")
            references_in_board.update(
                [
                    job.get("reference")
                    for job in response["data"]
                    if job.get("reference")
                ]
            )
            # FIXME why + 1 ? max_page + 1 in get_all_job_pages_from_board
            if page == response["meta"]["maxPage"] + 1:
                break
            page += 1

        references_to_archive = references_in_board - references_to_push
        adapter.info(
            "Sync mode enabled. Archiving {} items from board".format(
                len(references_to_archive)
            )
        )
        for reference in references_to_archive:
            response = hrflow_client.job.indexing.archive(
                board_key=parameters.board_key, reference=reference, is_archive=1
            )
            if response["code"] >= 400:
                adapter.error(
                    "Failed to archive job in board board_key={} "
                    "reference={} response={}".format(
                        parameters.board_key, reference, response
                    )
                )
                raise Exception("Failed to archive job")
        adapter.info("Archiving finished")

    for job in jobs:
        reference = job.get("reference")
        if reference is None:
            if parameters.enrich_with_parsing:
                adapter.info("Starting parsing for job without reference")
                try:
                    enrich_job_with_parsing(hrflow_client, job)
                    adapter.info("Parsing finished")
                except JobParsingException as e:
                    adapter.error(
                        "Failed to parse job response={}".format(e.client_response)
                    )
                    failed_jobs.append(job)
                    continue
            response = hrflow_client.job.indexing.add_json(
                board_key=parameters.board_key, job_json=job
            )
            if response["code"] >= 400:
                adapter.error(
                    "Failed to index job with no reference "
                    "board_key={} response={}".format(parameters.board_key, response)
                )
                failed_jobs.append(job)
            continue

        response = hrflow_client.job.indexing.get(
            board_key=parameters.board_key, reference=reference
        )
        if "Unable to find object: job" in response["message"]:
            if parameters.enrich_with_parsing:
                adapter.info(
                    "Starting parsing for job with reference={}".format(reference)
                )
                try:
                    enrich_job_with_parsing(hrflow_client, job)
                    adapter.info("Parsing finished")
                except JobParsingException as e:
                    adapter.error(
                        "Failed to parse job response={}".format(e.client_response)
                    )
                    failed_jobs.append(job)
                    continue
            response = hrflow_client.job.indexing.add_json(
                board_key=parameters.board_key, job_json=job
            )
            if response["code"] >= 400:
                adapter.error(
                    "Failed to index job board_key={} reference={} response={}".format(
                        parameters.board_key, reference, response
                    )
                )
                failed_jobs.append(job)
                continue
        elif response["code"] == 200:
            archived_at = response["data"].get("archived_at")
            job_key = response["data"]["key"]
            if archived_at is None:
                if parameters.update_content:
                    response = hrflow_client.job.indexing.edit(
                        board_key=parameters.board_key, key=job_key, job_json=job
                    )
                    if response["code"] >= 400:
                        adapter.error(
                            "Failed to edit job board_key={} "
                            "reference={} response={}".format(
                                parameters.board_key, reference, response
                            )
                        )
                        failed_jobs.append(job)
                        continue
            else:
                response = hrflow_client.job.indexing.archive(
                    board_key=parameters.board_key, reference=reference, is_archive=0
                )
                if response["code"] >= 400:
                    adapter.error(
                        "Failed to unarchive job board_key={} "
                        "reference={} response={}".format(
                            parameters.board_key, reference, response
                        )
                    )
                    failed_jobs.append(job)
                    continue
                response = hrflow_client.job.indexing.edit(
                    board_key=parameters.board_key, key=job_key, job_json=job
                )
                if response["code"] >= 400:
                    adapter.error(
                        "Failed to edit job board_key={} "
                        "reference={} response={}".format(
                            parameters.board_key, reference, response
                        )
                    )
                    failed_jobs.append(job)
                    continue
        else:
            adapter.error(
                "Failed to get job from board board_key={} "
                "reference={} response={}".format(
                    parameters.board_key, reference, response
                )
            )
            failed_jobs.append(job)
            continue

    return failed_jobs


HrFlowJobWarehouse = Warehouse(
    name="HrFlow.ai Jobs",
    data_schema=HrFlowJob,
    data_type=DataType.job,
    write=WarehouseWriteAction(parameters=WriteJobParameters, function=write),
)
